# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.detection.spot_modeling module.
"""

# TODO test bigfish.detection.fit_subpixel
# TODO test bigfish.detection.build_reference_spot
# TODO test bigfish.detection.modelize_spot
# TODO test bigfish.detection.precompute_erf
