# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.detection.spot_detection module.
"""

# TODO test bigfish.detection.detect_spots
# TODO test bigfish.detection.local_maximum_detection
# TODO test bigfish.detection.spots_thresholding
# TODO test bigfish.detection.automated_threshold_setting
