# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for bigfish.segmentation.nuc_segmentation module.
"""


# TODO add test for bigfish.segmentation.remove_segmented_nuc
